# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
""" Copyright (c) 2000-2003 LOGILAB S.A. (Paris, FRANCE).
 http://www.logilab.fr/ -- mailto:contact@logilab.fr

 basic checker for Python code
"""

__revision__ = "$Id: classes.py,v 1.35 2004/02/02 13:00:14 syt Exp $"

import inspect

from logilab.common import astng

from logilab.pylint.interfaces import IASTNGChecker
from logilab.pylint.checkers import BaseChecker, CheckerHandler
from logilab.pylint.checkers.utils import is_abstract, is_interface, \
     is_exception, get_nodes_from_class


MSGS = {
    'F0201': ('Unable to check method %r of interface %s',
              'Used when PyLint has been unable to fetch a method declared in \
              an interface (either in the class or in the interface) and so to\
              check its implementation.'),
    'F0202': ('Unable to check method (%s)',
              'Used when PyLint has been unable to check a method for an \
              unexpected raison. Please report this kind of error.'),
    'F0203': ('Unable to resolve (%s)',
              'Used when PyLint has been unable to resolve a name.'),

    'E0201': ('Access to undefined member %r',
              'Used when an instance member not defined in the instante, its\
              class or its ancestors is accessed.'),
    'E0202': ('Method hide an inherited attribute from %s',
              'Used when a class defines a method which hide an attribute from \
              an ancestor class.'),
    'W0201': ('Attribute %r defined outside __init__',
              'Used when an instance attribute is defined outside the __init__\
              method.'),
    
    'E0211': ('Method has no argument',
              'Used when a method which should have "self" or "cls" (for class \
              method) as argument has no argument defined.'),
    'E0212': ('Class method should have "cls" as first argument',
              'Used when a class method has an attribute different than "cls" as\
              first argument, to easily differentiate them from regular \
              instance methods.'),
    'E0213': ('Method doesn\'t have "self" as first argument',
              'Used when a method has an attribute different the "self" as\
              first argument.'),
    'W0211': ('Static method with %r as first argument',
              'Used when a static method has "self" or "cls" as first argument.'
              ),
    'E0221': ('Interface %s is not a class (%s)',
              'Used when a class claims to implement an interface which is not \
              a class.'),
    'E0222': ('Missing method %r from %s interface' ,
              'Used when a method declared in an interface is missing from a \
              class implementing this interface'),
    'W0221': ('Arguments number differs from %s method',
              'Used when a method has a different number of arguments than in the \
              implemented interface or in an overriden method.'),
    'W0222': ('Signature differs from %s method',
              'Used when a method signature is different than in the \
              implemented interface or in an overriden method.'),
    'W0223': ('Method %r is abstract in class %r but is not overriden',
              'Used when an abstract method (ie raise NotImplementedError) is \
              not overriden in concrete class.'
              ),
    
    'W0231': ('__init__ method from base class %r is not called',
              'Used when an ancestor class method has an __init__ method \
              which is not called by a derived class.'),
    'W0232': ('Class has no __init__ method',
              'Used when a class has no __init__ method, neither its parent \
              classes.'),
    'W0233': ('__init__ method from a non direct base class %r is called',
              'Used when an __init__ method is called on a class which is not \
              in the direct ancestors for the analysed class.'),
    
    }


class ClassChecker(BaseChecker, CheckerHandler):
    """checks for :                                                            
    * methods without self as first argument                                   
    * overriden methods signature                                              
    * access only to existant members via self                                 
    * attributes not defined in the __init__ method                            
    * supported interfaces implementation                                      
    * unreachable code                                                         
    """
    
    __implements__ = (IASTNGChecker,)

    # configuration section name
    name = 'classes'
    # messages
    msgs = MSGS
    priority = -2
    # configuration options
    options = (('ignore-iface-methods',
                {'default' : ('isImplementedBy', 'deferred', 'extends', 'names',
                              'namesAndDescriptions', 'queryDescriptionFor', 
                              'getBases', 'getDescriptionFor', 'getDoc',
                              'getName', 'getTaggedValue', 'getTaggedValueTags',
                              'isEqualOrExtendedBy', 'setTaggedValue',
                              'isImplementedByInstancesOf'),
                 'type' : 'csv',
                 'metavar' : '<method names>',
                 'help' : 'List of interface methods to ignore, \
separated by a comma. This is used for instance to not check methods defines \
in Zope\'s Interface base class.'}
                ),
               ('ignore-mixin-members',
                {'default' : 1, 'type' : 'yn', 'metavar': '<y_or_n>',
                 'help' : 'Tells wether missing members accessed in mixin \
class should be ignored. A mixin class is detected if its name ends with \
"mixin" (case insensitive).'}
                ),
               )

    def visit_class(self, node):
        """init visit variable _accessed and check interfaces
        """
        self._accessed = {}
        # FIXME: should not depend on object
        if node.object is not None:
            self.check_bases_classes(node)
            self.check_interfaces(node)
            if not (is_interface(node.object) or is_exception(node.object)):
                try:
                    node.get_method('__init__')
                except astng.NotFoundError:
                    self.add_message('W0232', args=node, node=node)

    def leave_class(self, class_node):
        """close a class node :
        check that instance attributes are defined in __init__ and check
        access to existant members
        """
        # checks attributes are defined in __init__
        for attr, node in class_node.instance_attrs.items():
            frame = node.get_frame()
            if frame.name not in ('__init__', '__new__', 'setUp'):
                try:
                    par_node = class_node.get_ancestor_for_attribute(attr)
                    frame = par_node.instance_attrs[attr].get_frame()
                    if frame.name not in ('__init__', '__new__',
                                          'setUp'):
                        self.add_message('W0201', args=attr, node=node)
                except astng.NotFoundError:
                    self.add_message('W0201', args=attr, node=node)
        # check access to existant members
        if not self.config.ignore_mixin_members or \
               class_node.name[-5:].lower() != 'mixin':
            self.check_accessed_members(class_node)
        del self._accessed
            
    def visit_function(self, node):
        """check method arguments, overriding
        """
        # check first argument is self if this is actually a method
        if node.is_method():
            self.check_first_arg_for_type(node)
            if node.name == '__init__':
                self.check_init(node)
                return
            klass = node.parent.get_frame()
            # check signature if the method overrload an herited method
            try:
                overriden = klass.get_ancestor_for_method(node.name)
            except astng.NotFoundError:
                pass
            except astng.ASTNGBuildingException:
                return
            else:
                # get astng for the searched method
                try:
                    meth_node = overriden.locals[node.name]
                    self.check_signature(node, meth_node, 'overriden')
                except KeyError:
                    # we have found the method but it's not in the local
                    # dictionnary.
                    # This may happen with astng build from living objects only
                    pass
            # check if the method overload an attribute
            try:
                overriden = klass.get_ancestor_for_attribute(node.name)
                self.add_message('E0202', args=overriden.name, node=node)
            except astng.NotFoundError:
                pass
                
    def leave_function(self, node):
        """check method arguments, overriding
        """
        # check first argument is self if this is actually a method
        if node.is_method():
            del self._first_attr
            return

    def visit_getattr(self, node):
        """check if the name handle an access to a class member
        if so, register it
        """
        if hasattr(self, '_first_attr') and isinstance(node.expr, astng.Name):
            if node.expr.name == self._first_attr:
                self._accessed.setdefault(node.attrname, []).append(node)

                
    def check_accessed_members(self, node):
        """check that accessed members are defined
        """
        for attr, nodes in self._accessed.items():
            # is it a builtin attribute
            if attr in ('__dict__', '__class__'):
                continue
            # is it an instance attribute ?
            if node.instance_attrs.has_key(attr):
                continue
            # or a class attribute ?
            if node.locals.has_key(attr):
                continue
            # or an inherited method / attribute ?
            try:
                node.get_ancestor_for_method(attr)
            except astng.NotFoundError:
                pass
            except astng.ASTNGBuildingException:
                continue
            else:
                continue
            try:
                node.get_ancestor_for_attribute(attr)
            except astng.NotFoundError:
                pass
            else:
                continue
            for _node in nodes:
                self.add_message('E0201', node=_node, args=attr)
        
    def check_first_arg_for_type(self, node):
        """check the name of first argument, expect:
        'self' for a regular method
        'cls' for a class method
        not 'self' nor 'cls' for a static method
        """
        self._first_attr = node.argnames and node.argnames[0]
        # static method
        if node.is_static_method():
            if node.argnames and node.argnames[0] in ('self', 'cls'):
                self.add_message('W0211', args=node.argnames[0], node=node)
            self._first_attr = None
        # class / regular method with no args
        elif not node.argnames:
            self.add_message('E0211', node=node)
        # class method
        elif node.is_class_method():
            if self._first_attr != 'cls':
                self.add_message('E0212', node=node)
        # regular method without self as argument
        elif self._first_attr != 'self':
            self.add_message('E0213', node=node)

    def check_bases_classes(self, node):
        """check that the given class node implements abstract methods from
        base classes
        """
        obj = node.object
        for name in dir(obj):
            try:
                member = getattr(obj, name)
            except AttributeError:
                # occurs on ExtensionClass.Base !
                continue
            if not inspect.ismethod(member):
                continue
            if inspect.isbuiltin(member.im_func):
                continue
            # get class method astng
            try:
                method_node = self._get_method(node, name)
            except astng.NotFoundError:
            # FIXME : should not need try:except and test on method_node
            # (usually means we need a fix to astng.resolve...)
                continue
            if method_node is None:
                continue
            # ignore method defined in the class
            meth_class = method_node.parent.get_frame()
            if meth_class is node:
                continue
            # check that the ancestor's method is not abstract
            if is_abstract(method_node, pass_is_abstract=0):
                self.add_message('W0223', node=node,
                                 args=(name, meth_class.name))
                    
    def check_interfaces(self, node):
        """check that the given class node really implements declared
        interfaces
        """
        try:
            implements = node.object.__implements__
        except AttributeError:
            return
        manager = self.linter.manager
        if not type(implements) in (type(()), type([])):
            implements = (implements,)
        for iface in implements:
            if not inspect.isclass(iface):
                self.add_message('E0221', node=node,
                                 args=(iface, type(iface)))
                continue
            iface_node = manager.astng_from_class(iface)
            for name in dir(iface):
                if name.startswith('_') or \
                       name in self.config.ignore_iface_methods:
                    # don't check method begining with an underscore, usually
                    # belonging to the interface implementation
                    continue
                member = getattr(iface, name)
                if not inspect.ismethod(member):
                    continue
                # get class method astng
                try:
                    method_astng = self._get_method(node, name)
                except astng.NotFoundError:
                    self.add_message('E0222', args=(name, iface_node.name),
                                     node=node)
                    continue
                if method_astng is None:
                    continue
                # don't go further if it is an inherited method
                if not method_astng.parent.get_frame() is node:
                    continue
                # get iface method astng
                imethod_astng = self._get_method(iface_node, name)
                if imethod_astng is None:
                    continue
                # check signature
                self.check_signature(method_astng, imethod_astng,
                                     '%s interface' % iface_node.name)

    def check_init(self, node):
        """check that the __init__ method call ancestor __init__
        """
        to_call = {}
        unresolved = {}
        klass_node = node.parent.get_frame()
        for base in klass_node.basenames:
            parts = base.split('.')
            cbase = parts[0]
            try:
                baseobj = klass_node.resolve(cbase)
            except (astng.ResolveError, astng.NotFoundError):
                unresolved[base] = 1
                self.add_message('F0203', node=node, args=cbase)
                continue
            for part in parts[1:]:
                try:
                    cbase = '%s.%s' % (cbase, part)
                    baseobj = getattr(baseobj, part)
                except AttributeError:
                    unresolved[base] = 1
                    self.add_message('F0203', node=node, args=cbase)
            else:
                if hasattr(baseobj, '__init__'):
                    to_call[base] = 1
        for stmt in get_nodes_from_class(node, astng.CallFunc):
            expr = stmt.node
            if not (isinstance(expr, astng.Name) or
                    isinstance(expr, astng.Getattr)):
                continue
            func_str = expr.as_string()
            if func_str.endswith('.__init__'):
                klass_name = '.'.join(func_str.split('.')[:-1])
                try:
                    del to_call[klass_name]
                except KeyError:
                    if unresolved.has_key(klass_name):
                        continue
                    self.add_message('W0233', node=expr, args=klass_name)
        for klass_name in to_call.keys():
            self.add_message('W0231', args=klass_name, node=node)
            
    def check_signature(self, method1, method2, class_type):
        """check that the signature of the two given methods match
        
        class_type is in 'class', 'interface'
        """
        if not isinstance(method1, astng.Function):
            self.add_message('F0222', args=method1, node=method1)
            return
        if not isinstance(method2, astng.Function):
            self.add_message('F0222', args=method2, node=method1)
            return
        if len(method1.argnames) != len(method2.argnames):
            self.add_message('W0221', args=class_type, node=method1)
        elif len(method1.defaults) != len(method2.defaults):
            self.add_message('W0222', args=class_type, node=method1)
            
    def _get_method(self, node, method_name, fatal=1):
        """get astng for <method_name> on the given class node
        """
        try:
            return node.get_method(method_name)
        except astng.NotFoundError:
            raise
        except Exception:
            if fatal:
                self.add_message('F0201', args=(method_name, node.name),
                                 node=node)
#            import traceback
#            traceback.print_exc()
        
        
def register(linter):
    """required method to auto register this checker """
    linter.register_checker(ClassChecker(linter))
